//
// © 2016 David Vines
//
// Creative Commons License
//
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
//
// Space filling curve used for simplicification - we will assume two adjacent addresses in this list are adjacent
// Used for determining neighbours
/* globals ROT, Game */

function HouseState(adjective,condition) {
    this._adjective = adjective;
    this._condition = condition;
    this._worse =  this;
    this._better = this;
}

HouseState.prototype = {
    _setBetter: function(better) {
        this._better = better || this;
    },
    _setWorse: function(worse) {
        this._worse = worse || this;
    },
    improved: function() {
        return this._better;
    },
    worsened: function() {
        return this._worse;
    },
    getAdjective: function() {
        return this._adjective;
    },
    getCondition: function() {
        return this._condition;
    },
};

HouseState._states = [new HouseState("dilapidated",40),
                      new HouseState("run-down",25),
                      new HouseState("in need of work",15),
                      new HouseState("faded",5),
                      new HouseState("tidy",-5),
                      new HouseState("handsome",-15),
                      new HouseState("beautiful",-30),
                      new HouseState("immaculate",-45)
];
(function() {
    for(var i=0; i<HouseState._states.length; i++) {
        if (i !== HouseState._states.length-1 ) {
            HouseState._states[i]._setBetter(HouseState._states[i+1]);
        }
        if (i !== 0) {
            HouseState._states[i]._setWorse(HouseState._states[i-1]);
        }
    }
}());

HouseState.forAdjective = function(adj) {
    for(var i=0; i<HouseState._states.length; i++) {
        if (HouseState._states[i].getAdjective() === adj) {
            return HouseState._states[i];
        }
    }
    throw new Error("No House State for adjective '"+adj+"'.");
};
HouseState.randomState = function() {
    var weights = {"dilapidated":7,
            "run-down":5,
            "in need of work":5,
            "faded":3,
            "tidy":3,
            "handsome":6,
            "beautiful":6,
            "immaculate":5
    };
    return HouseState.forAdjective(ROT.RNG.getWeightedValue(weights));
};
HouseState.renovatedState = function() {
    var weights = {"handsome":18,
            "beautiful":12,
            "immaculate":5
    };
    return HouseState.forAdjective(ROT.RNG.getWeightedValue(weights));
};

function Address(index,address,coords) {
    this._index = index;
    this._address = address;
    this._people = [];
    this._state = HouseState.randomState();
    this._coords = coords;
}

Address.prototype = {
    getAddress: function() {
        return this._address;
    },
    getInhabitants: function() {
        return this._people;
    },
    getHouseholder: function() {
        return this._people[0];
    },
    addInhabitant: function(person) {
        this._people.push(person);
    },
    forEachInhabitant: function(f) {
        this._people.forEach(f);
    },
    getNeighbours: function() {
        var left  = (this._index === 0 ? Address._addresses.length-1 : this._index-1);
        var right = (this._index === Address._addresses.length-1 ? 0 : this._index+1);
        return [ Address._addresses[left], Address._addresses[right] ];
    },
    getHouseState: function() {
        return this._state;
    },
    renovate: function() {
        this._state = HouseState.renovatedState();
        if (this === Address.getSelected()) {
            Address.displayData(this)();
        }
    },
    nextYear: function() {
        if (ROT.RNG.getPercentage() < 10) {
            this._state = this._state.worsened();
        } else if (ROT.RNG.getPercentage() < 10) {
            this._state = this._state.improved();
        }
    },
    getCoords: function() {
        return this._coords;
    }
};

Address.selectedChangeListener = undefined;
Address.displayData = function(address) {
    return function() {
        var html = "<h6>"+address.getAddress()+"</h6>";
        html += "<p>The paintwork of "+address.getAddress()+" is "+address.getHouseState().getAdjective()+".</p><p>The householder is ";
        html += address.getInhabitants()[0].description(Game._electorate,Game._factors,",")+"</p>";
        if (address.getInhabitants().length > 1) {
            if (address.getInhabitants()[0].getName === "YOU") {
                html += "<p>You recognise ";
            } else {
                html += "<p>"+address.getInhabitants()[0].She()+" recognises ";
            }
            html += "the following as electors for the ward of Oldgate:<ul>";
            html += address.getInhabitants().slice(1).reduce(function(string,elector){
                return string+"<li>"+elector.description(Game._electorate,Game._factors,",")+"</li>";
            },"");
            html += "</ul></p>";
        }
        Game.setIdHTML("map.description",html);
        if (Address._selected !== address) {
            Address._selected = address;
            Address.redrawCanvas();
            if (Address.selectedChangeListener) {
                Address.selectedChangeListener(address);
            }
        }
    };
};

Address.clickAddress = function(address) {
    return function() {
        // Mouseover equivalent for this address (for touch devices)
        Address.displayData(address)();
        // Disable mouseover
        Address.activateMapEventHandlers(false,undefined,undefined);
        
        // Until the mouse leaves the image
        var img = document.getElementById("map.img");
        img.onmouseout = Game.captureExceptionEventHandler(function(e) {
            e = e || event;
            var isOutsideMap = function(e) {
                var map = document.getElementById("map.map");
                var elem = e.relatedTarget || e.toElement || e.fromElement;
                while(elem && elem !== map) {
                    elem = elem.parentNode;
                }
                return (elem !== map);
            };
            if (isOutsideMap(e)) {
                img.onmouseout = undefined;
                Address.activateMapEventHandlers(true,undefined,undefined);
            }
        });
    };
};

Address.selectedClickListener = undefined;
Address.doubleClickAddress = function(address) {
    return function() {
        if (Address.selectedClickListener) {
            Address.selectedClickListener();
        }
    };
};

Address._addresses = [];
Address._selected = undefined;
Address.addAddress = function(address,coords) {
    var index = Address._addresses.length;
    var a = new Address(index,address,coords);
    Address._addresses.push(a);
    return a;
};
Address.getAddresses = function() {
    return Address._addresses;
};
Address.forEach = function(f) {
    Address._addresses.forEach(f);
};
Address.random = function() {
    return Address._addresses.random();
};

Address.getSelected = function() {
    return Address._selected;
};

Address.redrawCanvas = function() {
    // Get the size of the map and resize the canvas and map to it.
    var img = document.getElementById("map.img");
    document.getElementById("map.data").style.height = img.clientHeight+"px";
    var canvas = document.getElementById("map.canvas");
    canvas.width = img.clientWidth;
    canvas.height = img.clientHeight;
    var context = canvas.getContext("2d");
    context.linewidth = 4;
    context.strokeStyle = '#008';
    context.fillStyle = '#008';
    var area = document.getElementById("map.map").firstElementChild;
    Address.forEach(function(a) {
        var coords = a.getCoords();
        var newcoords = [];
        context.beginPath();
        newcoords[0] = coords[0]*canvas.width/640;
        newcoords[1] = coords[1]*canvas.height/489;
        context.moveTo(newcoords[0],newcoords[1]);
        for(var i=2; i < coords.length; i+=2) {
            newcoords[i] = coords[i]*canvas.width/640;
            newcoords[i+1] = coords[i+1]*canvas.height/489;
            context.lineTo(newcoords[i],newcoords[i+1]);
        }
        context.closePath();
        if (a === Address._selected) {
            context.fill();
        }
        context.stroke();
        area.coords = newcoords.join();
        area = area.nextElementSibling;
    });
};

Address.activateMapEventHandlers = function(mouseover,click,dblclick) {
    var area = document.getElementById("map.map").firstElementChild;
    Address.forEach(function(a) {
        if (mouseover !== undefined) {
            area.onmouseover = (mouseover ? Game.captureExceptionFunction(Address.displayData(a)) : undefined);
        }
        if (click !== undefined) {
            area.onclick = (click ? Game.captureExceptionFunction(Address.clickAddress(a)) : undefined);
        }
        if (dblclick !== undefined) {
            area.ondblclick = (dblclick ? Game.captureExceptionFunction(Address.doubleClickAddress(a)) : undefined);
        }
        area = area.nextElementSibling;
    });
};

Address.activateMap = function(flag) {
    document.getElementById("map.random").onclick = (flag ? function() {
        Address.displayData(Address.random())();
    } : undefined);
    document.getElementById("map.random").disabled = !flag;
    if (!flag) {
        document.getElementById("map.img").onmouseout = undefined;
    }
    Address.activateMapEventHandlers(flag,flag,flag);
};

// Basic image size = 640 x 489
Address.init = function() {
    var img = document.getElementById("map.map");
    for(var i=0; i<img.children.length; i++) {
        var addr = img.children[i].alt;
        var coordsAsText = img.children[i].coords.split(',');
        var coords = [];
        for(var j=0; j<coordsAsText.length; j++)  {
            coords[j] = parseInt(coordsAsText[j]);
        }
        Address.addAddress(addr,coords);
    }
    document.getElementById("map.random").disabled = true;
};