//
// © 2016 David Vines
//
// Creative Commons License
//
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
//
/* globals Game, Elector, module, exports */

function Reason(code,shortDescription,description,implement,longReason,score,reply,needToBeCouncillor) {
    needToBeCouncillor = needToBeCouncillor || false;
    this._code = code;
    this._shortDescription = shortDescription;
    this._description = description;
    this._needToBeCouncillor = needToBeCouncillor;
    this._longReason = longReason;

    if (typeof implement !== "function") {
        throw new Error("The implementation '"+implement+"' passed on construction of '"+code+"' is not a function.");
    }
    if (typeof score !== "function") {
        throw new Error("The score '"+score+"' passed on construction of '"+code+"' is not a function.");
    }
    if (typeof reply !== "function") {
        throw new Error("The reply '"+reply+"' passed on construction of '"+code+"' is not a function.");
    }

    this._implement = implement;
    this._score = score;
    this._reply = reply;
    this._scoreDeltaForAddress = Reason._defaultScoreDeltaForAddress;
}

Reason.prototype = {
    getCode: function() {
        return this._code;
    },
    getDescription: function() {
        return this._description;
    },
    getShortDescription: function() {
        return this._shortDescription;
    },
    needsCouncillor: function() {
        return this._needToBeCouncillor;
    },
    implement: function(topic,view,person,score) {
        return this._implement(topic,view,person,score);
    },
    getLongReason: function() {
        return this._longReason;
    },
    getScore: function(line) {
        return this._score(line);
    },
    getReply: function(person,score) {
        return this._reply(person,score);
    },
    scoreDeltaForAddress: function(address) {
        return this._scoreDeltaForAddress(address);
    },
    _setScoreDeltaForAddress: function(f) {
        if (typeof f !== "function") {
            throw new Error("The scoreDeltaForAddress '"+f+"' passed to '"+this._code+"' is not a function.");
        }
        this._scoreDeltaForAddress = f;
    }
};

Reason._reasons = {};

Reason.forCode = function(code) {
    if (Reason.isValidCode(code)) {
        return Reason._reasons[code];
    } else {
        throw new Error("Unknown reason code '"+code+"'");
    }
};

Reason.getReasons = function(isCouncillor) {
    var rs = [];
    for(var p in Reason._reasons) {
        if (Reason._reasons.hasOwnProperty(p)) {
            if (!Reason._reasons[p].needsCouncillor() || isCouncillor) {
                rs.push(Reason._reasons[p]);
            }
        }
    }
    return rs;
};

Reason.isValidCode = function(code) {
    return Reason._reasons.hasOwnProperty(code);
};

Reason.addNewReason = function(code,shortdescription,description,implement,longReason,score,reply,needToBeCouncillor) {
    Reason._reasons[code] = new Reason(code,shortdescription,description,implement,longReason,score,reply,needToBeCouncillor);
};

Reason._defaultScoreDeltaForAddress = function(address) {
    return 0;
};
Reason._workScoreDeltaForAddress = function(address) {
    return address.getHouseState().getCondition();
};

Reason.addNewReason("work","painting the house","Concentrate on your painting business letting your customers know you're hoping to be elected (your customers may change their vote to you)",Game.implementWorkHard,"You meet with %p to renew the paintwork of %a, which is %h.",Elector.scoreWork,Elector.replyWork);
Reason.addNewReason("fun","having a good time","Concentrate on having fun, but trading drinks for votes as you go (your drinking buddies may change their vote for you)",Game.implementPlayHard,"You meet with %p in their local pub.",Elector.scoreFun,Elector.replyFun);
Reason.addNewReason("tell","voting","Concentrate on persuading those who support you to actually vote (your voters may be more likely to vote)",Game.implementTell,"You meet with %p to encourage them to vote.",Elector.scoreTell,Elector.replyTell);
Reason.addNewReason("campaign","voting for you","Campaign hard to get elected (some people may change their vote to you)",Game.implementCampaign,"You meet with %p at their home, %a, to persuade them to vote for you.",Elector.scoreCampaign,Elector.replyCampaign);
Reason.addNewReason("smear","voting against the alderman","Campaign against the alderman",Game.implementSmear,"You meet with %p at their home, %a, to persuade them to not vote for %d.",Elector.scoreSmear,Elector.replySmear);
Reason.addNewReason("watch","the local watch","Become active in the ward watch (people helped by the watch may change their vote to you)",Game.implementWatch,"You meet with %p and end up discussing a recent burglary.",Elector.scoreWatch,Elector.replyWatch);
Reason.addNewReason("sherrif","the alderman","Work to get the current alderman elected as sherrif or even Lord of the City (The current alderman may be elected to sherrif (and hence be ineligble to be elected as a councillor or alderman)",Game.implementPromote,"You work at the Witanmoot to get %d elected as Sherrif.",Elector.scorePromote,Elector.replyPromote,true);
Reason.addNewReason("wardmoot","the wardmoot","Become active in the wardmoot (people who didn't vote may change their vote to you)",Game.implementWardMoot,"You meet with %p at their home, %a, so you can present their concerns at the wardmoot.",Elector.scoreWardmoot,Elector.replyWardmoot,true);
Reason.addNewReason("witanmoot","the Witanmoot","Become active in the Witanmoot (people who actually voted last time may change their vote to you)",Game.implementWitanmoot,"You meet with %p at their home, %a, so you can present their concerns at the Witanmoot.",Elector.scoreWitanmoot,Elector.replyWitanmoot,true);

Reason.forCode("work")._setScoreDeltaForAddress(Reason._workScoreDeltaForAddress);

// For testing via QUnit
if (typeof module !== "undefined" && module.exports) {
    exports.Reason = Reason;
}